import os
import subprocess
from db import get_prompt_key, get_stream_key, r_db, request_stack_key
from llama_cpp import Llama
import json
from loguru import logger

llama_model_path = os.environ.get(
    "LLAMA_MODEL_PATH", "models/mistral-7b-instruct-v0.1.Q4_K_M.gguf"
)
download_url = os.environ.get(
    "LLAMA_MODEL_DOWNLOAD_URL",
    "https://huggingface.co/TheBloke/Mistral-7B-Instruct-v0.1-GGUF/resolve/main/mistral-7b-instruct-v0.1.Q4_K_M.gguf",
)

# Download if necessary
# model=
if not os.path.exists(llama_model_path):
    logger.info(f"Downloading Llama model to {llama_model_path}")
    os.makedirs(os.path.dirname(llama_model_path), exist_ok=True)
    res = subprocess.run(
        [
            "wget",
            "-q",
            download_url,
            "-O",
            llama_model_path,
        ]
    )
else:
    logger.info(f"Using Llama model at {llama_model_path}")

_llm = Llama(
    model_path=llama_model_path,
    n_ctx=8192,
    n_gpu_layers=80,
    verbose=False,
    n_batch=512,
)


def llm(query, temperature):
    response = _llm(
        query,
        max_tokens=8,
        temperature=temperature,
        stop=["</s>"],
    )
    return response


def listen(*, request_queue_name):
    logger.info("Pinging Redis")
    response = r_db.ping()
    logger.info(f"Redis ping response: {response}")

    logger.info("Warming up Llama model")
    _llm("Hello")
    logger.info(f"Listening for requests on {request_queue_name}")
    while True:
        item = r_db.blpop(request_queue_name)
        match item:
            case (bytes(_), bytes(request_id)):
                request_id = request_id.decode("utf-8")
                try:
                    process_request(request_queue_name, request_id)
                except Exception as e:
                    logger.error(
                        f"Error processing request {request_id}: {e}. Won't retry."
                    )
            case _:
                logger.warning(f"Received unknown item {item}")


def process_request(request_queue_name, request_id):
    logger.info(f"Processing request {request_id}")
    prompt_key = get_prompt_key(request_id)
    query = r_db.get(prompt_key)
    if query is None:
        logger.warning(f"Prompt {prompt_key} not found. " "Is the key expired? ")
        return

    query = json.loads(query)
    query_str = query['prompt']
    temperature = query['temperature']
    response = llm(query_str, temperature=temperature)
    # Put the response in the stream
    # This will be picked up by the client
    r_db.rpush(get_stream_key(request_id), json.dumps(response))
    if response["choices"][0]["finish_reason"] != "stop":
        # Update the prompt, so the next request in the queue can continue
        query['prompt'] += response["choices"][0]["text"]
        r_db.set(prompt_key, json.dumps(query))
        # update the expiry time
        r_db.expire(prompt_key, 10)
        if r_db.llen(request_queue_name) == 0:
            # If there are no more requests in the queue, process this one again
            # This is to make use of kv cache in the transformer
            process_request(request_queue_name, request_id)
        else:
            r_db.rpush(request_queue_name, request_id)


if __name__ == "__main__":
    listen(request_queue_name=request_stack_key)
